package owee

  1. Overview
  2. Docs

Module type Owee_marker.T0Source

The signature that needs to be implemented to add service support to your objects.

Sourcetype t
Sourceval service : t -> 'result service -> 'result service_result

The service dispatch function. Given a value of type t and a service, try to satisfy this service.